// Persistence of Vision Ray Tracer Scene Description File
// File: rolling.inc
// Vers: 3.6.1
// Desc: Macros to generate rolling-matrices 
// Date: 05.08.05 (dd.mm.yy)
// Auth: Tim Nikias Wenclawiak
// Last Update: 21.08.05 (dd.mm.yy)

//Description
// Provides a set of Macros to create and calculate a proper
// rotation-matrix to simulate the rolling of a spherical object.
// Thus, it is best used with spheres. ;-)
// All it takes is a radius and a series of positions, which are
// provided on a frame-by-frame basis, and based upon the radius
// and the distance (and an optional up-vector), the axis of
// rotation and the amount is calculated.

//The Macros
//==========

// RollMatrix(X,Y,Z,Pos)
//  -Creates a matrix from the given vectors

//Create_RollFile(Filename,Radius)
//  -Creates a rollfile

//Move_RollFile(Filename,Pos)
//  -Translates the Rollfile to the given Position *without* rolling effect

//RollFile(Filename)
//  -Loads the Rollfile and returns a transformation

//Roll2Position(Filename,Pos,Ground_Normal)
//  -Rolls the given file to the new Position, using the Ground_Normal for calculations

//RollAlongArray(Filename,Array)
//  -Works like a succession of Roll2Positions. If the Array is 1-dimensional, y is used
//   as Ground_Normal, otherwise, the 2nd dimension of the array will be used.

//Randomize_RollFile(Filename)
//  -Randomizes the initial orientation of the matrix

//Internal Macros, don't use these:

//Load_RollFile(Filename)
//  -Loads a Rollfile, for internal use only
//Save_RollFile(Filename,X,Y,Z,Pos,Rad)
//  -Saves a Rollfile, for internal use only
//Check_RollData(X,Y,Z,Pos,Radius)
//  -Checks the given data, internal use only

//Homepage:
// www.nolights.de
//Email:
// LOWERCASEONLYtimISNTnikias[AT]gmx.netWARE

//A macro to take care of the vnormalize(<0,0,0>)-stop, which can
//be a PITA in calculations...
#macro VNormalize(_Vec)
 #if (vlength(_Vec))
  #local _Ret=vnormalize(_Vec); #else
  #local _Ret=<0,0,0>; #end
 _Ret
#end

//Create_Matrix
//Function:
// Generates Matrix-Transform from given coordinate-vectors
//Parameters:
// 1. X-Axis
// 2. Y-Axis
// 3. Z-Axis
// 4. Position in normal 3D-Space
//Returns:
// Matrix-Transform
//Notes:
// None
#macro RollMatrix(_Xn,_Yn,_Zn,_Tn)
  #local _X = VNormalize(_Xn);
  #local _Y = VNormalize(_Yn);
  #local _Z = VNormalize(_Zn); 
  //Return
    matrix <_X.x,_X.y,_X.z,
            _Y.x,_Y.y,_Y.z,
            _Z.x,_Z.y,_Z.z,
            _Tn.x,_Tn.y,_Tn.z>
#end

//Create_RollFile
//Function:
// Writes default-Rollfile with Matrix-Vectors
//Parameters:
// 1. Filename
// 2. Radius
//Returns:
// File with 4 Vectors
//Notes:
// None
#macro Create_RollFile(_Filename,_Rad)
  #fopen _Oroll _Filename write
  #local _X = <1,0,0>;
  #local _Y = <0,1,0>;
  #local _Z = <0,0,1>;
  #local _T = <0,0,0>;
  #local _R = _Rad;
  #write (_Oroll, _X, ",\n")
  #write (_Oroll, _Y, ",\n")
  #write (_Oroll, _Z, ",\n")
  #write (_Oroll, _T, ",\n")
  #write (_Oroll, _R, "\n")
  #fclose _Oroll
#end

//Load_RollFile
//Loads a Rollfile, internal use only
#macro Load_RollFile(_Filename)
  #fopen _Iroll _Filename read
  #read (_Iroll, _Xi, _Yi, _Zi, _Ti, _Ri)
  #fclose _Iroll
#end

//Save_RollFile
//Saves a RollFile, internal use only
#macro Save_RollFile(_Filename,_X,_Y,_Z,_T,_R)
  Check_RollData(_X,_Y,_Z,_T,_R)
  #fopen _Oroll _Filename write
  #write (_Oroll, _X, ",\n")
  #write (_Oroll, _Y, ",\n")
  #write (_Oroll, _Z, ",\n")
  #write (_Oroll, _T, ",\n")
  #write (_Oroll, _R, "\n")
  #fclose _Oroll
#end

//Move_RollFile
//Moves the center without rolling
#macro Move_RollFile(_Filename,_Pos)
  Load_RollFile(_Filename)
  Save_RollFile(_Filename,_Xi,_Yi,_Zi,_Pos,_Ri)
  #undef _Xi
  #undef _Yi
  #undef _Zi
  #undef _Ti
  #undef _Ri
#end

//Check_RollData
//Checks the matrix for perpendicularity and validity
#macro Check_RollData(_X,_Y,_Z,_T,_R)
  #if (vlength(_X))
    #declare _X = vnormalize(_X);
    #local _Xcheck = true;
  #else
    #local _Xcheck = false;
  #end
  #if (vlength(_Y))
    #declare _Y = vnormalize(_Y);
    #local _Ycheck = true;
  #else
    #local _Ycheck = false;
  #end
  #if (vlength(_Z))
    #declare _Z = vnormalize(_Z);
    #local _Zcheck = true;
  #else
    #local _Zcheck = false;
  #end
  #if ((!_Xcheck & !_Ycheck) | (!_Xcheck & !_Zcheck) | (!_Ycheck & !_Zcheck))
    #declare _X = x;
    #declare _Y = y;
    #declare _Z = z;  
  #end
  #if (_R=0)
    #declare _R=1;  
  #end
  #local _Axis = 1;
  #if (_Xcheck) #local _Axis = _Axis*2; #end
  #if (_Ycheck) #local _Axis = _Axis*3; #end
  #if (_Zcheck) #local _Axis = _Axis*5; #end
  //Check on which axis to base the recalibration
  #if (mod(_Axis,2)=0)
    #local _xA = _X;
    #if (mod(_Axis,3)=0)
      #local _xB = _Y;
    #else
      #local _xB = _Z;
    #end
  #else
    #local _xA = _Y;
    #local _xB = _Z;
  #end
  //Recalibrate
  #local _xC = vcross(_xA,_xB);
  #local _xB = vcross(_xC,_xA);
  //Return calibrated matrix
  #if (mod(_Axis,2)=0)
    #local _X = _xA;
    #if (mod(_Axis,3)=0)
      #local _Y = _xB;
      #local _Z = _xC;
    #else
      #local _Y = _xC;
      #local _Z = _xB;
    #end
  #else
    #local _Y = _xA;
    #local _Z = _xB;
    #local _X = _xC;
  #end
#end                                     

//RollFile
//Function:
// Loads a RollFile and converts it to a matrix
//Parameters:
// 1. Filename
//Returns:
// 1. Matrix-Transform for centered object
//Notes:
// None
#macro RollFile2Transform(_Filename)
  Load_RollFile(_Filename)
  transform{RollMatrix(_Xi,_Yi,_Zi,_Ti)}
  #undef _Xi
  #undef _Yi
  #undef _Zi
  #undef _Ti
  #undef _Ri
#end

//Roll2Position
//Function:
// Rolls the data of the file to the given position,
// using the given up-vector for rotation-reference
//Parameters:
// 1. Filename to load and use
// 2. Position to roll to
// 3. Up-Vector
//Returns:
// Modified File
//Notes:
// None
#macro Roll2Position(_Filename,_Position,_Up)
  Load_RollFile(_Filename)
  #local _Relative = _Position-_Ti;
  #if (vlength(_Relative)!=0 & vdot(VNormalize(_Relative),vnormalize(_Up))!=1)
    #local _Distance = vlength(_Relative);
    #local _RollAmount = 360*_Distance/(2*pi*_Ri);
    #local _RollAxis = vcross(_Up,vnormalize(_Relative));
    #local _Xn = vaxis_rotate(_Xi,_RollAxis,_RollAmount);
    #local _Yn = vaxis_rotate(_Yi,_RollAxis,_RollAmount);
    #local _Zn = vaxis_rotate(_Zi,_RollAxis,_RollAmount);
  #else
    #local _Xn = _Xi;
    #local _Yn = _Yi;
    #local _Zn = _Zi;
  #end
  #local _Tn = _Position;
  Save_RollFile(_Filename,_Xn,_Yn,_Zn,_Tn,_Ri)
  #undef _Xi
  #undef _Yi
  #undef _Zi
  #undef _Ti
  #undef _Ri
#end

//RollAlongArray(_Filename,_PosArray)
//Function:
// Rolls the data from position to position. If the array
// is 2-dimensional, the second dimension is used as the
// up-vector for the rotation, otherwise, y is assumed.
//Parameters:
// 1. Filename to load and use
// 2. Positions [and up-vectors in 2nd dimension]
//Returns:
// Modified File
//Notes:
// None
#macro RollAlongArray(_Filename,_PosArray)
  Load_RollFile(_Filename)
  #local dim = dimensions(_PosArray);
  #local dim_size = dimension_size(_PosArray,1);
  //1-Dimensional, y is up
  #if (dim = 1)
    #local _dimCounter = 0;
    #while (_dimCounter < dim_size)
      #local _Position = _PosArray[_dimCounter];
      #local _Relative = _Position-_Ti;
      #if (vlength(_Relative)!=0 & vdot(VNormalize(_Relative),y)!=1)
        #local _Distance = vlength(_Relative);
        #local _RollAmount = 360*_Distance/(2*pi*_Ri);
        #local _RollAxis = vcross(y,vnormalize(_Relative));
        #local _Xn = vaxis_rotate(_Xi,_RollAxis,_RollAmount);
        #local _Yn = vaxis_rotate(_Yi,_RollAxis,_RollAmount);
        #local _Zn = vaxis_rotate(_Zi,_RollAxis,_RollAmount);
      #else
        #local _Xn = _Xi;
        #local _Yn = _Yi;
        #local _Zn = _Zi;
      #end
      #local _Tn = _Position;
      #local _Xi = _Xn;
      #local _Yi = _Yn;
      #local _Zi = _Zn;
      #local _Ti = _Tn;
      #local _dimCounter = _dimCounter + 1;
    #end
  #end
  //2-Dimensions
  #if (dim = 2)
    #local _dimCounter = 0;
    #while (_dimCounter < dim_size)
      #local _Up = VNormalize(_PosArray[_dimCounter][1]);
      #if (vlength(_Up)!=1) #local _Up = y; #end
      #local _Position = _PosArray[_dimCounter][0];
      #local _Relative = _Position-_Ti;
      #if (vlength(_Relative)!=0 & vdot(VNormalize(_Relative),_Up)!=1)
        #local _Distance = vlength(_Relative);
        #local _RollAmount = 360*_Distance/(2*pi*_Ri);
        #local _RollAxis = vcross(_Up,vnormalize(_Relative));
        #local _Xn = vaxis_rotate(_Xi,_RollAxis,_RollAmount);
        #local _Yn = vaxis_rotate(_Yi,_RollAxis,_RollAmount);
        #local _Zn = vaxis_rotate(_Zi,_RollAxis,_RollAmount);
      #else
        #local _Xn = _Xi;
        #local _Yn = _Yi;
        #local _Zn = _Zi;
      #end
      #local _Tn = _Position;
      #local _Xi = _Xn;
      #local _Yi = _Yn;
      #local _Zi = _Zn;
      #local _Ti = _Tn;
      #local _dimCounter = _dimCounter + 1;
    #end
  #end
  Save_RollFile(_Filename,_Xi,_Yi,_Zi,_Ti,_Ri)
  #undef _Xi
  #undef _Yi
  #undef _Zi
  #undef _Ti
  #undef _Ri
#end

//Randomize_RollFile
//Function:
// Generates a random axis to start off with
//Parameters:
// 1. Filename
// 2. Seed
//Returns:
// Modified RollFile
//Notes:
// None
#macro Randomize_RollFile(_Filename,_Seed)
  #local _R=seed(_Seed);
  #local _X = vnormalize(<rand(_R),rand(_R),rand(_R)>*2-1);
  #local _Y = vnormalize(<rand(_R),rand(_R),rand(_R)>*2-1);
  #while (vlength(_X-_Y)=0)
    #local _Y = vnormalize(<rand(_R),rand(_R),rand(_R)>*2-1);
  #end
  #local _Z = vnormalize(vcross(_X,_Y));
  #local _Y = vnormalize(vcross(_X,_Z));
  Load_RollFile(_Filename)
  Save_RollFile(_Filename,_X,_Y,_Z,_Ti,_Ri)
  #undef _Xi
  #undef _Yi
  #undef _Zi
  #undef _Ti
  #undef _Ri
#end
